# -*- coding: utf-8 -*-
#
# Helper routines for working w/dask
#

# Builtin/3rd party package imports
import os
import sys
import socket
import subprocess
import getpass
import time
import inspect
import textwrap
import numpy as np
from tqdm import tqdm
if sys.platform == "win32":
    # tqdm breaks term colors on Windows - fix that (tqdm issue #446)
    import colorama
    colorama.deinit()
    colorama.init(strip=False)

# Local imports: differentiate b/w being imported as Syncopy sub-package or
# standalone ACME module: if imported by Syncopy, use some lambda magic to avoid
# circular imports due to (at import-time) only partially initialized Syncopy
from .shared import user_input, user_yesno
if "syncopy" in sys.modules:
    isSpyModule = True
    import syncopy as spy
    customIOError = lambda msg : spy.shared.errors.SPYIOError(msg)
    customValueError = lambda legal=None, varname=None, actual=None : \
        spy.shared.errors.SPYValueError(legal=legal, varname=varname, actual=actual)
    customTypeError = lambda val, varname=None, expected=None : \
        spy.shared.errors.SPYTypeError(val, varname=varname, expected=expected)
    scalar_parser = lambda var, varname="", ntype=None, lims=None : \
        spy.shared.parsers.scalar_parser(var, varname=varname, ntype=ntype, lims=lims)
else:
    isSpyModule = False
    from warnings import showwarning
    import logging
    from .shared import _scalar_parser as scalar_parser
    customIOError = IOError
    customValueError = lambda legal=None, varname=None, actual=None : ValueError(legal)
    customTypeError = lambda msg, varname=None, expected=None : TypeError(msg)

from dask_jobqueue import SLURMCluster
from dask.distributed import Client, get_client
from datetime import datetime, timedelta

# Be optimistic: prepare success message to be used throughout this module
_successMsg = "{name:s} Cluster dashboard accessible at {dash:s}"

__all__ = ["esi_cluster_setup", "local_cluster_setup", "cluster_cleanup", "slurm_cluster_setup"]


# Setup SLURM jobs on the ESI HPC cluster
def esi_cluster_setup(partition="8GBXS", n_jobs=2, mem_per_job="auto", n_jobs_startup=100,
                      timeout=60, interactive=True, interactive_wait=120, start_client=True,
                      job_extra=[], **kwargs):
    """
    Start a Dask distributed SLURM worker cluster on the ESI HPC infrastructure
    (or local multi-processing)

    Parameters
    ----------
    partition : str
        Name of SLURM partition/queue to start workers in. Use the command `sinfo`
        in the terminal to see a list of available SLURM partitions on the ESI HPC
        cluster.
    n_jobs : int
        Number of SLURM jobs (=workers) to spawn
    mem_per_job : None or str
        Memory booking for each job. Can be specified either in megabytes
        (e.g., ``mem_per_job = 1500MB``) or gigabytes (e.g., ``mem_per_job = "2GB"``).
        If `mem_per_job` is `None`, or `"auto"` it is attempted to infer a sane default value
        from the chosen partition, e.g., for ``partition = "8GBS"`` `mem_per_job` is
        automatically set to the allowed maximum of `'8GB'`. However, even in
        queues with guaranted memory bookings, it is possible to allocate less
        memory than the allowed maximum per job to spawn numerous low-memory
        jobs. See Examples for details.
    n_jobs_startup : int
        Number of spawned jobs to wait for. If `n_jobs_startup` is `100` (default),
        the code does not proceed until either 100 SLURM jobs are running or the
        `timeout` interval has been exceeded.
    timeout : int
        Number of seconds to wait for requested jobs to start (see `n_jobs_startup`).
    interactive : bool
        If `True`, user input is queried in case not enough jobs (set by `n_jobs_startup`)
        could be started in the provided waiting period (determined by `timeout`).
        The code waits `interactive_wait` seconds for a user choice - if none is
        provided, it continues with the current number of running jobs (if greater
        than zero). If `interactive` is `False` and no job could not be started
        within `timeout` seconds, a `TimeoutError` is raised.
    interactive_wait : int
        Countdown interval (seconds) to wait for a user response in case fewer than
        `n_jobs_startup` jobs could be started. If no choice is provided within
        the given time, the code automatically proceeds with the current number of
        active dask workers.
    start_client : bool
        If `True`, a distributed computing client is launched and attached to
        the dask worker cluster. If `start_client` is `False`, only a distributed
        computing cluster is started to which compute-clients can connect.
    job_extra : list
        Extra sbatch parameters to pass to SLURMCluster.
    **kwargs : dict
        Additional keyword arguments can be used to control job-submission details.

    Returns
    -------
    proc : object
        A distributed computing client (if ``start_client = True``) or
        a distributed computing cluster (otherwise).

    Examples
    --------
    The following command launches 10 SLURM jobs with 2 gigabytes memory each
    in the `8GBS` partition

    >>> client = esi_cluster_setup(n_jobs=10, partition="8GBS", mem_per_job="2GB")

    The underlying distributed computing cluster can be accessed using

    >>> client.cluster

    Notes
    -----
    The employed parallel computing engine relies on the concurrent processing library
    `Dask <https://docs.dask.org/en/latest/>`_. Thus, the distributed computing
    clients generated here are in fact instances of :class:`dask.distributed.Client`.
    This function specifically acts  as a wrapper for :class:`dask_jobqueue.SLURMCluster`.
    Users familiar with Dask in general and its distributed scheduler and cluster
    objects in particular, may leverage Dask's entire API to fine-tune parallel
    processing jobs to their liking (if wanted).

    See also
    --------
    slurm_cluster_setup : start a distributed Dask cluster of parallel processing workers using SLURM
    local_cluster_setup : start a local Dask multi-processing cluster on the host machine
    cluster_cleanup : remove dangling parallel processing job-clusters
    """

    # Re-direct printing/warnings to ACME logger outside of SyNCoPy
    customPrint, customWarning = _logging_setup()

    # For later reference: dynamically fetch name of current function
    funcName = "{pre:s}<{pkg:s}{name:s}>".format(pre="Syncopy " if isSpyModule else "",
                                                 pkg="ACME: " if isSpyModule else "",
                                                 name=inspect.currentframe().f_code.co_name)

    # Check if SLURM's `sinfo` can be accessed
    proc = subprocess.Popen("sinfo",
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                            text=True, shell=True)
    _, err = proc.communicate()

    # Any non-zero return-code means SLURM is not ready to use
    if proc.returncode != 0:

        # SLURM is not installed: either allocate `LocalCluster` or just leave
        if proc.returncode > 0:
            if interactive:
                msg = "{name:s} SLURM does not seem to be installed on this machine " +\
                    "({host:s}). Do you want to start a local multi-processing " +\
                    "computing client instead? "
                startLocal = user_yesno(msg.format(name=funcName, host=socket.gethostname()),
                                        default="no")
            else:
                startLocal = True
            if startLocal:
                return local_cluster_setup(interactive=interactive, start_client=start_client)

        # SLURM is installed, but something's wrong
        msg = "{preamble:s}SLURM queuing system from node {node:s}. " +\
              "Original error message below:\n{error:s}"
        raise customIOError(msg.format(preamble=funcName + " Cannot access " if not isSpyModule else "",
                                       node=socket.gethostname(),
                                       error=err))

    # Use default by-job worker count or extract it from anonymous keyword args (if provided)
    workers_per_job = 1
    if kwargs.get("workers_per_job") is not None:
        workers_per_job = kwargs.pop("workers_per_job")

    # Extract by-job CPU core count from anonymous keyword args or...
    if kwargs.get("n_cores") is not None:
        n_cores = kwargs.pop("n_cores")
    else:
        # ...get memory limit (*in MB*) of chosen partition and set core count
        # accordingly (multiple of 8 wrt to GB RAM)
        try:
            pc = subprocess.run("scontrol -o show partition {}".format(partition),
                                capture_output=True, check=True, shell=True, text=True)
            defMem = int(pc.stdout.strip().partition("DefMemPerCPU=")[-1].split()[0])
        except Exception as exc:
            msg = "{preamble:s}available memory per CPU in chosen SLURM partition. " +\
                "Original error message below:\n{error:s}"
            raise customIOError(msg.format(preamble=funcName + " Cannot access " if not isSpyModule else "",
                                           error=str(exc)))
        n_cores = int(defMem / 8000)

    # Determine if `job_extra`` is a list (this is also checked in `slurm_cluster_setup`,
    # but we may need to append to it, so ensure that's possible)
    if not isinstance(job_extra, list):
        msg = "{} `job_extra` has to be List, not {}"
        raise customTypeError(job_extra if isSpyModule else msg.format(funcName, str(job_extra)),
                              varname="job_extra",
                              expected="list")

    # If '--output' was not provided, append default output folder to `job_extra`
    if not any(option.startswith("--output") or option.startswith("-o") for option in job_extra):
        usr = getpass.getuser()
        slurm_wdir = "/cs/slurm/{usr:s}/{usr:s}_{date:s}"
        slurm_wdir = slurm_wdir.format(usr=usr,
                                       date=datetime.now().strftime('%Y%m%d-%H%M%S'))
        os.makedirs(slurm_wdir, exist_ok=True)
        out_files = os.path.join(slurm_wdir, "slurm-%j.out")
        job_extra.append("--output={}".format(out_files))

    # Let the SLURM-specific setup function do the rest (returns client or cluster)
    return slurm_cluster_setup(partition, n_cores, n_jobs, workers_per_job, mem_per_job,
                               n_jobs_startup, timeout, interactive, interactive_wait,
                               start_client, job_extra, invalid_partitions=["DEV", "PPC"], **kwargs)


# Setup SLURM cluster
def slurm_cluster_setup(partition, n_cores, n_jobs, workers_per_job, mem_per_job,
                        n_jobs_startup, timeout, interactive, interactive_wait,
                        start_client, job_extra, invalid_partitions=[]):
    """
    Start a distributed Dask cluster of parallel processing workers using SLURM

    **NOTE** If you are working on the ESI HPC cluster, please use
    :func:`~acme.esi_cluster_setup` instead!

    Parameters
    ----------
    partition : str
        Name of SLURM partition/queue to use
    n_cores : int
        Number of CPU cores per SLURM job
    n_jobs : int
        Number of SLURM jobs to spawn
    workers_per_job : int
        Number of processess (=dask workers) to use per job. Should be greater
        than one only if the chosen partition contains nodes that expose multiple
        cores per job.
    mem_per_job : str
        Memory allocation for each job
    n_jobs_startup : int
        Number of spawned jobs to wait for. The code does not return until either
        `n_jobs_startup` SLURM jobs are running or the `timeout` interval (see
        below) has been exceeded.
    timeout : int
        Number of seconds to wait for requested jobs to start (see `n_jobs_startup`).
    interactive : bool
        If `True`, user input is queried in case not enough jobs (set by `n_jobs_startup`)
        could be started in the provided waiting period (determined by `timeout`).
        The code waits `interactive_wait` seconds for a user choice - if none is
        provided, it continues with the current number of running jobs (if greater
        than zero). If `interactive` is `False` and no job could not be started
        within `timeout` seconds, a `TimeoutError` is raised.
    interactive_wait : int
        Countdown interval (seconds) to wait for a user response in case fewer than
        `n_jobs_startup` jobs could be started. If no choice is provided within
        the given time, the code automatically proceeds with the current number of
        active dask workers.
    start_client : bool
        If `True`, a distributed computing client is launched and attached to
        the dask worker cluster. If `start_client` is `False`, only a distributed
        computing cluster is started to which compute-clients can connect.
    job_extra : list
        Extra sbatch parameters to pass to SLURMCluster.
    invalid_partition : list
        List of partition names (strings) that are not available for launching
        dask workers.

    Returns
    -------
    proc : object
        A distributed computing client (if ``start_client = True``) or
        a distributed computing cluster (otherwise).

    See also
    --------
    esi_cluster_setup : start a SLURM worker cluster on the ESI HPC infrastructure
    local_cluster_setup : start a local Dask multi-processing cluster on the host machine
    cluster_cleanup : remove dangling parallel processing job-clusters
    """

    # Re-direct printing/warnings to ACME logger outside of SyNCoPy
    customPrint, customWarning = _logging_setup()

    # For later reference: dynamically fetch name of current function
    funcName = "{pre:s}<{pkg:s}{name:s}>".format(pre="Syncopy " if isSpyModule else "",
                                                 pkg="ACME: " if isSpyModule else "",
                                                 name=inspect.currentframe().f_code.co_name)

    # Retrieve all partitions currently available in SLURM
    proc = subprocess.Popen("sinfo -h -o %P",
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                            text=True, shell=True)
    out, err = proc.communicate()

    # Any non-zero return-code means SLURM is not ready to use
    if proc.returncode != 0:
        msg = "{preamble:s}SLURM queuing system from node {node:s}. " +\
              "Original error message below:\n{error:s}"
        raise customIOError(msg.format(preamble=funcName + " Cannot access " if not isSpyModule else "",
                                       node=socket.gethostname(),
                                       error=err))

    # Format subprocess shell output
    options = out.split()

    # Make sure we're in a valid partition
    if partition not in options:
        valid = list(set(options).difference(invalid_partitions))
        lgl = "'" + "or '".join(opt + "' " for opt in valid)
        msg = "{} Invalid partition selection {}, available SLURM partitions are {}"
        raise customValueError(legal=lgl if isSpyModule else msg.format(funcName, str(partition), lgl),
                               varname="partition",
                               actual=partition)

    # Parse job count
    try:
        scalar_parser(n_jobs, varname="n_jobs", ntype="int_like", lims=[1, np.inf])
    except Exception as exc:
        raise exc

    # Get requested memory per job
    if isinstance(mem_per_job, str):
        if mem_per_job == "auto":
            mem_per_job = None
    if mem_per_job is not None:
        msg = "{} `mem_per_job` has to be a valid memory specifier (e.g., '8GB', '12000MB'), not {}"
        lgl = "string representation of requested memory (e.g., '8GB', '12000MB')"
        if not isinstance(mem_per_job, str):
            raise customTypeError(mem_per_job if isSpyModule else msg.format(funcName, str(mem_per_job)),
                                  varname="mem_per_job",
                                  expected="string")
        if not any(szstr in mem_per_job for szstr in ["MB", "GB"]):
            raise customValueError(legal=lgl if isSpyModule else msg.format(funcName, str(mem_per_job)),
                                   varname="mem_per_job",
                                   actual=mem_per_job)
        memNumeric = mem_per_job.replace("MB","").replace("GB","")
        try:
            memVal = float(memNumeric)
        except:
            raise customValueError(legal=lgl if isSpyModule else msg.format(funcName, str(mem_per_job)),
                                   varname="mem_per_job",
                                   actual=mem_per_job)
        if memVal <= 0:
            raise customValueError(legal=lgl if isSpyModule else msg.format(funcName, str(mem_per_job)),
                                   varname="mem_per_job",
                                   actual=mem_per_job)

    # Parse job-waiter count
    try:
        scalar_parser(n_jobs_startup, varname="n_jobs_startup", ntype="int_like", lims=[0, np.inf])
    except Exception as exc:
        raise exc

    # Get memory limit (*in MB*) of chosen partition (guaranteed to exist, cf. above)
    pc = subprocess.run("scontrol -o show partition {}".format(partition),
                        capture_output=True, check=True, shell=True, text=True)
    try:
        mem_lim = int(pc.stdout.strip().partition("MaxMemPerCPU=")[-1].split()[0])
    except IndexError:
        try:
            mem_lim = int(pc.stdout.strip().partition("DefMemPerCPU=")[-1].split()[0])
        except IndexError:
            mem_lim = np.inf

    # Consolidate requested memory with chosen partition (or assign default memory)
    if mem_per_job is None:
        mem_per_job = str(mem_lim) + "MB"
    else:
        if "MB" in mem_per_job:
            mem_req = int(mem_per_job[:mem_per_job.find("MB")])
        else:
            mem_req = int(round(float(mem_per_job[:mem_per_job.find("GB")]) * 1000))
        if mem_req > mem_lim:
            msg = "{name:s}`mem_per_job` exceeds limit of {lim:d}GB for partition {par:s}. " +\
                "Capping memory at partition limit. "
            customWarning(msg.format(name=funcName + " " if not isSpyModule else "", lim=mem_lim, par=partition),
                          RuntimeWarning, __file__, inspect.currentframe().f_lineno)
            mem_per_job = str(int(mem_lim)) + "GB"

    # Parse requested timeout period
    try:
        scalar_parser(timeout, varname="timeout", ntype="int_like", lims=[1, np.inf])
    except Exception as exc:
        raise exc

    # Parse requested interactive waiting period
    try:
        scalar_parser(interactive_wait, varname="interactive_wait", ntype="int_like",
                      lims=[0, np.inf])
    except Exception as exc:
        raise exc

    # Determine if cluster allocation is happening interactively
    if not isinstance(interactive, bool):
        msg = "{} `interactive` has to be Boolean, not {}"
        raise customTypeError(interactive if isSpyModule else msg.format(funcName, str(interactive)),
                              varname="interactive",
                              expected="bool")

    # Determine if a dask client was requested
    if not isinstance(start_client, bool):
        msg = "{} `start_client` has to be Boolean, not {}"
        raise customTypeError(start_client if isSpyModule else msg.format(funcName, str(start_client)),
                              varname="start_client",
                              expected="bool")

    # Determine if job_extra is a list
    if not isinstance(job_extra, list):
        msg = "{} `job_extra` has to be List, not {}"
        raise customTypeError(job_extra if isSpyModule else msg.format(funcName, str(job_extra)),
                              varname="job_extra",
                              expected="list")

    # Determine if job_extra options are valid
    for option in job_extra:
        msg = "{} `job_extra` has to be a valid sbatch option, not {}"
        if not isinstance(option, str):
            raise customTypeError(option if isSpyModule else msg.format(funcName, str(option)),
                                  varname="option",
                                  expected="string")
        if not option[0] == "-":
            lgl = "job_extra options should be flagged with - or --"
            raise customValueError(legal=lgl, varname="option", actual=option)

    # Ensure validity of requested worker threads
    try:
        scalar_parser(workers_per_job, varname="workers_per_job",
                      ntype="int_like", lims=[1, 16])
    except Exception as exc:
        raise exc

    # Check for sanity of requested core count
    try:
        scalar_parser(n_cores, varname="n_cores",
                      ntype="int_like", lims=[1, np.inf])
    except Exception as exc:
        raise exc

    # Check validity of '--output' option if provided
    userOutSpec = [option.startswith("--output") or option.startswith("-o") for option in job_extra]
    if any(userOutSpec):
        userOut = job_extra[userOutSpec.index(True)]
        outSpec = userOut.split("=")
        if len(outSpec) != 2:
            lgl = "the SLURM output directory must be specified using -o/--output=/path/to/file"
            raise customValueError(legal=lgl if isSpyModule else "{} {}, not {}".format(funcName, lgl, userOut),
                                   varname="job_extra",
                                   actual=userOut)
        slurm_wdir = os.path.split(outSpec[1])[0]
        if len(slurm_wdir) > 0:
            if isSpyModule:
                try:
                    spy.shared.parsers.io_parser(slurm_wdir, varname="job_extra", isfile=False)
                except Exception as exc:
                    raise exc
            else:
                msg = "{} `slurmWorkingDirectory` has to be an existing directory, not {}"
                if not os.path.isdir(os.path.expanduser(slurm_wdir)):
                    raise ValueError(msg.format(funcName, str(slurm_wdir)))

    # Create `SLURMCluster` object using provided parameters
    cluster = SLURMCluster(cores=n_cores,
                           memory=mem_per_job,
                           processes=workers_per_job,
                           local_directory=slurm_wdir,
                           queue=partition,
                           python=sys.executable,
                           header_skip=["-t", "--mem"],
                           job_extra=job_extra)
                           # interface="asdf", # interface is set via `psutil.net_if_addrs()`

    # Compute total no. of workers and up-scale cluster accordingly
    total_workers = n_jobs * workers_per_job
    worker_count = min(total_workers, n_jobs_startup)
    if worker_count < total_workers:
        # cluster.adapt(minimum=worker_count, maximum=total_workers)
        cluster.scale(total_workers)
        msg = "{} Requested job-count {} exceeds `n_jobs_startup`: " +\
            "waiting for {} jobs to come online, then proceed"
        customPrint(msg.format(funcName, total_workers, n_jobs_startup))
    else:
        cluster.scale(total_workers)

    # Fire up waiting routine to avoid unfinished cluster setups
    if _cluster_waiter(cluster, funcName, worker_count, timeout, interactive, interactive_wait):
        return

    # Kill a zombie cluster in non-interactive mode
    if not interactive and _count_running_workers(cluster) == 0:
        cluster.close()
        err = "SLURM jobs could not be started within given time-out " +\
              "interval of {0:d} seconds"
        raise TimeoutError(err.format(timeout))

    # Highlight how to connect to dask performance monitor
    customPrint(_successMsg.format(name=funcName, dash=cluster.dashboard_link))

    # If client was requested, return that instead of the created cluster
    if start_client:
        return Client(cluster)
    return cluster


def _cluster_waiter(cluster, funcName, total_workers, timeout, interactive, interactive_wait):
    """
    Local helper that can be called recursively
    """

    # Re-direct printing/warnings to ACME logger outside SyNCoPy
    customPrint, _ = _logging_setup()

    # Wait until all workers have been started successfully or we run out of time
    wrkrs = _count_running_workers(cluster)
    to = str(timedelta(seconds=timeout))[2:]
    fmt = "{desc}: {n}/{total} \t[elapsed time {elapsed} | timeout at " + to + "]"
    ani = tqdm(desc="{} SLURM workers ready".format(funcName), total=total_workers,
               leave=True, bar_format=fmt, initial=wrkrs, position=0)
    counter = 0
    while _count_running_workers(cluster) < total_workers and counter < timeout:
        time.sleep(1)
        counter += 1
        ani.update(max(0, _count_running_workers(cluster) - wrkrs))
        wrkrs = _count_running_workers(cluster)
        ani.refresh()   # force refresh to display elapsed time every second
    ani.close()

    # If we ran out of time before all workers could be started, ask what to do
    if counter == timeout and interactive:
        msg = "{name:s} SLURM jobs could not be started within given time-out " +\
              "interval of {time:d} seconds"
        customPrint(msg.format(name=funcName, time=timeout))
        query = "{name:s} Do you want to [k]eep waiting for 60s, [a]bort or " +\
                "[c]ontinue with {wrk:d} workers?"
        choice = user_input(query.format(name=funcName, wrk=wrkrs),
                            valid=["k", "a", "c"], default="c", timeout=interactive_wait)

        if choice == "k":
            return _cluster_waiter(cluster, funcName, total_workers, 60, True, 60)
        elif choice == "a":
            customPrint("{} Closing cluster...".format(funcName))
            cluster.close()
            return True
        else:
            if wrkrs == 0:
                query = "{} Cannot continue with 0 workers. Do you want to " +\
                        "[k]eep waiting for 60s or [a]bort?"
                choice = user_input(query.format(funcName), valid=["k", "a"],
                                    default="a", timeout=60)
                if choice == "k":
                    _cluster_waiter(cluster, funcName, total_workers, 60, True, 60)
                else:
                    customPrint("{} Closing cluster...".format(funcName))
                    cluster.close()
                    return True

    return False


def local_cluster_setup(interactive=True, start_client=True):
    """
    Start a local distributed Dask multi-processing cluster

    Parameters
    ----------
    interactive : bool
        If `True`, a confirmation dialog is displayed to ensure proper encapsulation
        of calls to `local_cluster_setup` inside a script's main module block.
        See Notes for details. If `interactive` is `False`, the dialog is not shown.
    start_client : bool
        If `True`, a distributed computing client is launched and attached to
        the workers. If `start_client` is `False`, only a distributed
        computing cluster is started to which compute-clients can connect.

    Returns
    -------
    proc : object
        A distributed computing client (if ``start_client = True``) or
        a distributed computing cluster (otherwise).

    Notes
    -----
    The way Python spawns new processes requires an explicit separation of initialization
    code (i.e., code blocks that should only be executed once) from the actual
    program code. Specifically, everything that is supposed to be invoked only
    once by the parent spawner must be encapsulated in a script's main block.
    Otherwise, initialization code is not only run once by the parent process but
    executed by every child process at import time.

    This means that starting a local multi-processing cluster *has* to be wrapped
    inside a script's main module block, otherwise, every child process created by
    the multi-processing cluster starts a multi-processing cluster itself and so
    on escalating to an infinite recursion. Thus, if `local_cluster_setup` is
    called inside a script, it has to be encapsulated in the script's main module
    block, i.e.,

    .. code-block:: python

        if __name__ == "__main__":
            ...
            local_cluster_setup()
            ...

    Note that this capsulation is **only** required inside Python scripts. Launching
    `local_cluster_setup` in the (i)Python shell or inside a Jupyter notebook
    does not suffer from this problem.
    A more in-depth technical discussion of this limitation can be found in
    `Dask's GitHub issue tracker <https://github.com/dask/distributed/issues/2520>`_.

    Examples
    --------
    The following command launches a local distributed computing cluster using
    all CPU cores available on the host machine

    >>> client = local_cluster_setup()

    The underlying distributed computing cluster can be accessed using

    >>> client.cluster

    See also
    --------
    esi_cluster_setup : Start a distributed Dask cluster using SLURM
    cluster_cleanup : remove dangling parallel processing job-clusters
    """

    # Re-direct printing/warnings to ACME logger outside of SyNCoPy
    customPrint, _ = _logging_setup()

    # For later reference: dynamically fetch name of current function
    funcName = "{pre:s}<{pkg:s}{name:s}>".format(pre="Syncopy " if isSpyModule else "",
                                                 pkg="ACME: " if isSpyModule else "",
                                                 name=inspect.currentframe().f_code.co_name)

    # Determine if cluster allocation is happening interactively
    if not isinstance(interactive, bool):
        msg = "{} `interactive` has to be Boolean, not {}"
        raise customTypeError(interactive if isSpyModule else msg.format(funcName, str(interactive)),
                              varname="interactive",
                              expected="bool")

    # Determine if a dask client was requested
    if not isinstance(start_client, bool):
        msg = "{} `start_client` has to be Boolean, not {}"
        raise customTypeError(start_client if isSpyModule else msg.format(funcName, str(start_client)),
                              varname="start_client",
                              expected="bool")

    # Check, if we're running inside a Jupyter notebook...
    try:
        ipy = get_ipython()
        if ipy.__class__.__name__ == "ZMQInteractiveShell":
            maybeScript = False # Jupyter Notebook
        else:
            maybeScript = True  # iPython shell
    except NameError:
        maybeScript = True      # Python shell

    # ...if not, print warning/info message
    if maybeScript:
        msg = """\
        {name:s}If you use a script to start a local parallel computing client, please ensure
        the call to `local_cluster_setup` is wrapped inside a main module block, i.e.,

            if __name__ == "__main__":
                ...
                local_cluster_setup()
                ...

        Otherwise, a RuntimeError is raised due to an infinite recursion triggered by
        new processes being started before the calling process can finish its bootstrapping
        phase.
        """
        customPrint(textwrap.dedent(msg.format(name=funcName + " " if not isSpyModule else "")))

    # Additional safe-guard: if a script is executed, double-check with the user
    # for proper main idiom usage
    if interactive:
        msg = "{name:s}If launched from a script, did you wrap your code inside a main module block?"
        if not user_yesno(msg.format(name=funcName + " " if not isSpyModule else ""), default="no"):
            return

    # Start the actual distributed client
    client = Client()
    successMsg = "{name:s} Local parallel computing client ready. \n" + _successMsg
    customPrint(successMsg.format(name=funcName, dash=client.cluster.dashboard_link))
    if start_client:
        return client
    return client.cluster


def cluster_cleanup(client=None):
    """
    Stop and close dangling parallel processing workers

    Parameters
    ----------
    client : dask distributed computing client or None
        Either a concrete `dask client object <https://distributed.dask.org/en/latest/client.html>`_
        or `None`. If `None`, a global client is queried for and shut-down
        if found (without confirmation!).

    Returns
    -------
    Nothing : None

    See also
    --------
    esi_cluster_setup : Launch SLURM workers on the ESI compute cluster
    """

    # Re-direct printing/warnings to ACME logger outside SyNCoPy
    customPrint, customWarning = _logging_setup()

    # For later reference: dynamically fetch name of current function
    funcName = "{pre:s}<{pkg:s}{name:s}>".format(pre="Syncopy " if isSpyModule else "",
                                                 pkg="ACME: " if isSpyModule else "",
                                                 name=inspect.currentframe().f_code.co_name)

    # Attempt to establish connection to dask client
    if client is None:
        try:
            client = get_client()
        except ValueError:
            msg = "{name:s}No dangling clients or clusters found."
            customWarning(msg.format(name="" if isSpyModule else funcName + " "),
                          RuntimeWarning,
                          __file__,
                          inspect.currentframe().f_lineno)
            return
        except Exception as exc:
            raise exc
    else:
        if not isinstance(client, Client):
            msg = "{} `client` has to be a dask client object, not {}"
            customTypeError(client if isSpyModule else msg.format(funcName, str(client)),
                            varname="client",
                            expected="dask client object")

    # Prepare message for prompt
    if client.cluster.__class__.__name__ == "LocalCluster":
        userClust = "LocalCluster hosted on {}".format(client.scheduler_info()["address"])
    else:
        userName = getpass.getuser()
        outDir = client.cluster.job_header.partition("--output=")[-1]
        jobID = outDir.partition("{}_".format(userName))[-1].split(os.sep)[0]
        userClust = "cluster {0}_{1}".format(userName, jobID)
    nWorkers = len(client.cluster.workers)

    # If connection was successful, first close the client, then the cluster
    client.close()
    client.cluster.close()

    # Communicate what just happened and get outta here
    msg = "{fname:s} Successfully shut down {cname:s} containing {nj:d} workers"
    customPrint(msg.format(fname=funcName,
                     nj=nWorkers,
                     cname=userClust))

    return


def _count_running_workers(cluster):
    """
    Local replacement for the late `._count_active_workers` class method
    """
    return len(cluster.scheduler_info.get('workers'))


def _logging_setup():
    """
    Local helper to customize warning and print functionality at runtime
    """
    if isSpyModule:
        pFunc = print
        wFunc = lambda msg, kind, caller, lineno : spy.shared.errors.SPYWarning(msg, caller=caller)
    else:
        pFunc = print
        wFunc = showwarning
        allLoggers = list(logging.root.manager.loggerDict.keys())
        idxList = [allLoggers.index(loggerName) for loggerName in allLoggers \
            for moduleName in ["ACME", "ParallelMap"] if moduleName in loggerName]
        if len(idxList) > 0:
            logger = logging.getLogger(allLoggers[idxList[0]])
            pFunc = logger.info
            wFunc = lambda msg, kind, caller, lineno: logger.warning(msg)
    return pFunc, wFunc



# -*- coding: utf-8 -*-
#
# User-exposed interface of acme
#

# Builtin/3rd party package imports
import inspect
import sys
import numpy as np
import dask.array as da

# Local imports
from .backend import ACMEdaemon
from . import shared as acs
isSpyModule = False
if "syncopy" in sys.modules:
    isSpyModule = True

__all__ = ["ParallelMap"]


# Main context manager for parallel execution of user-defined functions
class ParallelMap(object):

    msgName = "{pre:s}<{pkg:s}ParallelMap>".format(pre="Syncopy " if isSpyModule else "",
                                                   pkg="ACME: " if isSpyModule else "")
    argv = None
    kwargv = None
    func = None
    n_inputs = None
    log = None
    _maxArgSize = 1024

    def __init__(
        self,
        func,
        *args,
        n_inputs="auto",
        write_worker_results=True,
        write_pickle=False,
        partition="auto",
        n_jobs="auto",
        mem_per_job="auto",
        setup_timeout=60,
        setup_interactive=True,
        stop_client="auto",
        verbose=None,
        dryrun=False,
        logfile=None,
        **kwargs):
        """
        Context manager that executes user-defined functions in parallel

        Parameters
        ----------
        func : callable
            User-defined function to be executed concurrently. Input arguments
            and return values should be "simple" (i.e., regular Python objects or
            NumPy arrays). See Notes for more information and Examples for
            details.
        args : arguments
            Positional arguments of `func`. Should be regular Python objects
            (lists, tuples, scalars, strings etc.) or NumPy arrays. See Notes
            for more information and Examples for details.
        kwargs : keyword arguments
            Keyword arguments of `func` (if any). Should be regular Python objects
            (lists, tuples, scalars, strings etc.) or NumPy arrays. See Notes
            for more information and Examples for details.
        n_inputs : int or "auto"
            Number of times `func` is supposed to be called in parallel. Usually,
            `n_inputs` does not have to be provided explicitly. If `n_inputs` is
            `"auto"` (default) this quantity is inferred from provided `args` and
            `kwargs`. This estimation may fail due to ambiguous input arguments
            (e.g., `args` and/or `kwargs` contain lists of differing lengths)
            triggering a `ValueError`. Only then is it required to set `n_input`
            manually. See Examples for details.
        write_worker_results : bool
            If `True`, the return value(s) of `func` is/are saved on disk (one
            HDF5 file per parallel worker). If `False`, the output of all parallel calls
            of `func` is collected in memory. See Examples and Notes for details.
        write_pickle : bool
            If `True`, the return value(s) of `func` is/are pickled to disk (one
            `'.pickle'`-file per parallel worker). Only effective if `write_worker_results`
            is `True`.
        partition : str
            Name of SLURM partition to use. If `"auto"` (default), the memory footprint
            of `func` is estimated using dry-run stubs based on randomly sampling
            provided `args` and `kwargs`. Estimated memory usage dictates queue
            auto-selection under the assumption of short run-times. For instance, on
            the ESI cluster a predicted memory footprint of 6 GB causes the `"8GBXS"`
            queue to be selected (minimal but sufficient memory and shortest runtime).
            To override auto-selection, provide name of SLURM queue explicitly. See, e.g.,
            :func:`~acme.esi_cluster_setup` for details.
        n_jobs : int or "auto"
            Number of SLURM jobs (=workers) to spawn. If `"auto"` (default), then
            ``n_jobs = n_inputs``, i.e., every SLURM worker performs a single
            call of `func`.
            If `n_inputs` is large and executing `func` is fast, setting
            ``n_jobs = int(n_inputs / 2)`` might be beneficial. See Notes for details.
        mem_per_job : str
            Memory booking for each SLURM worker. If `"auto"` (default), the standard
            value is inferred from the used partition (if possible). See, e.g.,
            :func:`~acme.esi_cluster_setup` for details.
        setup_timeout : int
            Timeout period (in seconds) for SLURM workers to come online. See, e.g.,
            :func:`~acme.esi_cluster_setup` for details.
        setup_interactive : bool
            If `True` (default), user input is queried in case not enough SLURM
            workers could be started within `setup_timeout` seconds. If no input
            is provided, the current number of spawned workers is used (even if
            smaller than the amount requested by `n_jobs`). If `False`, no user
            choice is requested.
        stop_client : bool or "auto"
            If `"auto"` (default), automatically started distributed computing clients
            are shut down at the end of computation, while user-provided clients
            are left untouched. If `False`, automatically started clients are
            left running after completion, user-provided clients are left untouched.
            If `True`, auto-generated clients *and* user-provided clients are
            shut down at the end of the computation.
        verbose : None or bool
            If `None` (default), general run-time information as well as warnings
            and errors are shown. If `True`, additionally debug information is
            shown. If `False`, only warnings and errors are propagated.
        dryrun : bool
            If `True` the user-provided function `func` is executed once using
            one of the input argument tuples prepared for the parallel workers (picked
            at random). If `setup_interactive` is `True`, a prompt asks if the
            actual parallel execution of `func` is supposed to be launched after the
            dry-run. The `dryrun` keyword is intended to test the correct
            distribution of arguments across workers prior to the actual concurrent
            computation.
        logfile : None or bool or str
            If `None` (default) or `False`, all run-time information as well as errors and
            warnings are printed to the command line only. If `True`, an auto-generated
            log-file is set up that records run-time progress. Alternatively, the
            name of a custom log-file can be provided (must not exist). The verbosity
            of recorded information can be controlled via setting `verbose`.

        Returns
        -------
        results : list
            If `write_worker_results` is `True`, `results` is a list of HDF5 file-names
            containing computed results. If `write_worker_results` is `False`,
            results is a list comprising the actual return values of `func`.


        .. _userguide:

        Examples
        --------
        Assume the function defined below is supposed to be run multiple times
        via SLURM for different values of `x`, `y` and `z`

        .. code-block:: python

            def f(x, y, z=3):
                return (x + y) * z

        The following code calls the function `f` with four different values of
        `x` (namely 2, 4, 6 and 8) setting `y` to 4 and leaving `z` at its default
        value of 3:

        .. code-block:: python

            from acme import ParallelMap

            with ParallelMap(f, [2, 4, 6, 8], 4) as pmap:
                results = pmap.compute()

        By default results are saved to disk in HDF5 format and `results` is a list
        of the corresponding filenames:

        >>> results
        ['/cs/home/username/ACME_20201217-135011-448825/f_0.h5',
         '/cs/home/username/ACME_20201217-135011-448825/f_1.h5',
         '/cs/home/username/ACME_20201217-135011-448825/f_2.h5',
         '/cs/home/username/ACME_20201217-135011-448825/f_3.h5']

        The contents of the containers can be accessed using `h5py`, e.g.,

        .. code-block:: python

            out = np.zeros((4, ))
            import h5py
            for ii, fname in enumerate(results):
                with h5py.File(fname, 'r') as f:
                    out[ii] = np.array(f['result_0'])

        which yields

        >>> out
        array([18., 24., 30., 36.])

        Alternatively, results may be collected directly in memory by setting
        `write_worker_results` to `False`. This is **not** recommended, since
        values have to be gathered from compute nodes via ethernet (slow) and
        are accumulated in the local memory of the interactive node you are using
        (potential memory overflow):

        .. code-block:: python

            with ParallelMap(f, [2, 4, 6, 8], 4, write_worker_results=False) as pmap:
                results = pmap.compute()

        Now `results` is a list of integers:

        >>> results
        [18, 24, 30, 36]

        Next, suppose `f` has to be evaluated for the same values of `x` (again
        2, 4, 6 and 8), but `y` is not a number but a NumPy array:

        .. code-block:: python

            y = np.ones((3,)) * 4
            with ParallelMap(f, [2, 4, 6, 8], y) as pmap:
                results = pmap.compute()

        This fails, because it is not clear which input is to be split up and distributed
        across workers for parallel execution:

        >>> ValueError: <ParallelMap> automatic input distribution failed: found 2 objects containing 3 to 4 elements. Please specify `n_inputs` manually.

        In this case, `n_inputs` has to be provided explicitly (`write_worker_results`
        is set to `False` for illustrative purposes only)

        .. code-block:: python

            with ParallelMap(f, [2, 4, 6, 8], y, n_inputs=4, write_worker_results=False) as pmap:
                results = pmap.compute()

        yielding

        >>> results
        [array([18., 18., 18.]),
         array([24., 24., 24.]),
         array([30., 30., 30.]),
         array([36., 36., 36.])]

        Now suppose `f` needs to be evaluated for fixed values of `x` and `y`
        with `z` varying randomly 500 times between 1 and 10. Since `f` is a
        very simple function, it is not necessary to spawn 500 SLURM jobs for this.
        Instead, allocate only 50 workers in the smallest available queue "8GBXS",
        i.e., each worker has to perform 10 evaluations of `f`. Additionally, keep the workers
        alive for re-use afterwards

        .. code-block:: python

            import numpy as np
            x = 2
            y = 4
            rng = np.random.default_rng()
            z = rng.integers(low=1, high=10, size=500, endpoint=True)

            with ParallelMap(f, x, y, z=z, n_jobs=50, partition="8GBXS", stop_client=False) as pmap:
                results = pmap.compute()

        This yields

        >>> len(results)
        500

        In a subsequent computation `f` needs to be evaluated for 1000 samples of
        `z`. In the previous call, `stop_client` was `False`, thus the next
        invocation of `ParallelMap` re-uses the existing SLURM worker swarm:

        .. code-block:: python

            z = rng.integers(low=1, high=10, size=1000, endpoint=True)

            with ParallelMap(f, x, y, z=z) as pmap:
                results = pmap.compute()

        Note the info message:

        >>> <ParallelMap> INFO: Attaching to global parallel computing client <Client: 'tcp://10.100.32.5:39747' processes=50 threads=50, memory=400.00 GB>

        Finally, suppose `f` has to be called for 20000 different values of `z`.
        Under the assumption that this computation takes a while, any run-time
        messages are to be written to a an auto-generated log-file:

        .. code-block:: python

            z = rng.integers(low=1, high=10, size=20000, endpoint=True)

            with ParallelMap(f, x, y, z=z, logfile=True) as pmap:
                results = pmap.compute()

        Alternatively, logging information may be written to a file my_log.txt instead

        .. code-block:: python

            z = rng.integers(low=1, high=10, size=20000, endpoint=True)

            with ParallelMap(f, x, y, z=z, logfile="my_log.txt") as pmap:
                results = pmap.compute()

        In some cases it might be necessary to work with objects that are not
        HDF5 compatible, e.g., sparse matrices created by `scipy.sparse`. Consider

        .. code-block:: python

            from scipy.sparse import spdiags

            ndim = 4
            x = spdiags(np.ones((ndim,)), 0, ndim, ndim)
            y = spdiags(3 * np.ones((ndim,)), 0, ndim, ndim)

        Then

        >>> x
        <4x4 sparse matrix of type '<class 'numpy.float64'>'
	            with 4 stored elements (1 diagonals) in DIAgonal format>
        >>> y
        <4x4 sparse matrix of type '<class 'numpy.float64'>'
	            with 4 stored elements (1 diagonals) in DIAgonal format>
        >>> x.toarray()
        array([[1., 0., 0., 0.],
               [0., 1., 0., 0.],
               [0., 0., 1., 0.],
               [0., 0., 0., 1.]])
        >>> y.toarray()
        array([[3., 0., 0., 0.],
               [0., 3., 0., 0.],
               [0., 0., 3., 0.],
               [0., 0., 0., 3.]])
        >>> f(x, y)
        <4x4 sparse matrix of type '<class 'numpy.float64'>'
	            with 4 stored elements (1 diagonals) in DIAgonal format>

        In this case, the default HDF5 storage format can be overridden using the
        keyword `write_pickle`

        .. code-block:: python

            with ParallelMap(f, x, y, n_inputs=5, write_pickle=True) as pmap:
                results = pmap.compute()

        which yields

        >>> results
        ['/cs/home/username/ACME_20201217-135011-448825/f_0.pickle',
         '/cs/home/username/ACME_20201217-135011-448825/f_1.pickle',
         '/cs/home/username/ACME_20201217-135011-448825/f_2.pickle',
         '/cs/home/username/ACME_20201217-135011-448825/f_3.pickle']

        Note that debugging programs running in parallel can be quite tricky.
        For instance, assume the function `f` is (erroneously) called with `z`
        set to `None`. In a regular sequential setting, identifying the problem
        is (relatively) straight-forward:

        >>> f(2, 4, z=None)
        TypeError: unsupported operand type(s) for *: 'int' and 'NoneType'

        However, when executing `f` in parallel using SLURM

        .. code-block:: python

            with ParallelMap(f, [2, 4, 6, 8], 4, z=None) as pmap:
                results = pmap.compute()

        the resulting error message can be somewhat overwhelming

        .. code-block:: python

            Function:  execute_task
            args:      ((<function reify at 0x7f425c25b0d0>, (<function map_chunk at 0x7f425c25b4c0>,
            <function ACMEdaemon.func_wrapper at 0x7f42569f1e50>, [[2], [4], [None], ['/cs/home/fuertingers/ACME_20201217-160137-984430'],
            ['f_0.h5'], [0], [<function f at 0x7f425c34bee0>]], ['z', 'outDir', 'outFile', 'taskID', 'userFunc'], {})))
            kwargs:    {}
            Exception: TypeError("unsupported operand type(s) for *: 'int' and 'NoneType'")
            slurmstepd: error: *** JOB 1873974 ON esi-svhpc18 CANCELLED AT 2020-12-17T16:01:43 ***

        To narrow down problems with parallel execution, the `compute` method
        of `ParallelMap` offers the `debug` keyword. If enabled, all function calls
        are performed in the local thread of the active Python interpreter. Thus, the execution
        is **not** actually performed in parallel. This allows regular error progration
        and even permits the use of tools like `pdb <https://docs.python.org/3/library/pdb.html>`_
        or ``%debug`` `iPython magics <https://ipython.readthedocs.io/en/stable/interactive/magics.html#magic-debug>`_.

        .. code-block:: python

            with ParallelMap(f, [2, 4, 6, 8], 4, z=None) as pmap:
                results = pmap.compute(debug=True)

        which results in

        .. code-block:: python

            <ipython-input-2-47feb885f020> in f(x, y, z)
                1 def f(x, y, z=3):
            ----> 2     return (x + y) * z

            TypeError: unsupported operand type(s) for *: 'int' and 'NoneType'

        In addition, the automaticall generated argument distribution to user-provided
        functions can be tested via the `dryrun` keyword. This permits to test-drive
        ACME's automatically generated argument lists prior to the actual concurrent
        computation, e.g.,

        >>> with ParallelMap(f, [2, 4, 6, 8], 4, dryrun=True) as pmap:
        >>>     results = pmap.compute()
        <ParallelMap> INFO: Performing a single dry-run of f simulating randomly picked worker #1 with automatically distributed arguments
        <ParallelMap> INFO: Dry-run completed. Elapsed time is 0.004725 seconds, estimated memory consumption was 0.01 MB.
        Do you want to continue executing f with the provided arguments? [Y/n] n

        In general it is strongly recommended to make sure any function supplied
        to `ParallelMap` works as intended in a sequential setting prior to running
        it in parallel.

        More examples and usage notes can be found in the package
        `README <https://github.com/esi-neuroscience/acme#acme-asynchronous-computing-made-easy>`_.

        Notes
        -----
        This code is solely intended for executing user-provided functions multiple
        times in parallel. Thus, only problems that can be split up into
        independent tasks can be processed with `ParallelMap` ("embarassingly parallel workloads").
        Inter-process communication, worker-synchronization or shared memory
        problems **will not work**.

        **User-Function Requirements**

        The user-provided function `func` has to meet some basic requirements to
        permit parallel execution with `ParallelMap`:

        * **input arguments of `func`** should be regular Python objects (lists, tuples,
          scalars, strings etc.) or NumPy arrays. Custom user-defined classes
          may or may not work. In general, anything that can be serialized via
          `cloudpickle <https://pypi.org/project/cloudpickle/>`_ should work out of the box.
        * if automatic result saving is used (`write_worker_results` is `True`),
          the **return value(s) of `func`** have to be suitable for storage in HDF5
          containers. Thus, anything returned by `func` should be either purely
          numeric (scalars or NumPy arrays) or purely lexical (strings). Hybrid
          text/numeric data-types (e.g., Pandas dataframes), custom class instances,
          functions, generators or complex objects (like matplotlib figures)
          **will not work**.

        **Auto-Generated HDF5-Files**

        All HDF5 files auto-generated by `ParallelMap` are stored in a directory
        *ACME_YYYYMMDD-hhmmss-ffffff* (encoding the current time as
        *YearMonthDay-HourMinuteSecond-Microsecond*) that is created in the user's
        home directory on ``hpx`` (if ACME is running on the ESI cluster) or the
        current working directory (if running locally). The HDF5 files themselves
        are named *funcname_workerid.h5*, where `funcname` is the name of the user-provided
        function and `workerid` encodes the number of the worker that generated
        the file (cf. Examples).

        The internal structure of all HDF5 files is kept as simple as possible:
        each return value of the user-provided function `func` is saved in a
        separate dataset in the file's root group. For instance, processing
        the following user-provided function

        .. code-block:: python

            def this_func(a, b, c):
                # ...some complicated calculations...
                return r0, r1, r2

        with 50 workers using ``write_worker_results = True`` yields 50 HDF5
        files *this_func_0.h5*, *this_func_1.h5*, ..., *this_func_49.h5* each
        containing three datasets `"result_0"` (holding `r0`), `"result_1"`
        (holding `r1`) and `"result_2"` (holding `r2`). User-provided functions
        with only a single return value correspondingly yield HDF5 files that
        only contain one dataset (`"result_0"`) in their respective root group.

        See also
        --------
        esi_cluster_setup : spawn custom SLURM worker clients on the ESI HPC cluster
        local_cluster_setup : start a local Dask multi-processing cluster on the host machine
        ACMEdaemon : Manager class performing the actual concurrent processing
        """

        # First and foremost, set up logging system (unless logger is already present)
        self.log = acs.prepare_log(func, caller=self.msgName, logfile=logfile, verbose=verbose)

        # Either guess `n_inputs` or use provided value to duplicate input args
        # and set class attributes `n_inputs`, `argv` and `kwargv`
        self.prepare_input(func, n_inputs, *args, **kwargs)

        # Create an instance of `ACMEdaemon` that does the actual parallel computing work
        self.daemon = ACMEdaemon(self,
                                 n_jobs=n_jobs,
                                 write_worker_results=write_worker_results,
                                 write_pickle=write_pickle,
                                 dryrun=dryrun,
                                 partition=partition,
                                 mem_per_job=mem_per_job,
                                 setup_timeout=setup_timeout,
                                 setup_interactive=setup_interactive,
                                 stop_client=stop_client)

    def prepare_input(self, func, n_inputs, *args, **kwargs):
        """
        User input parser

        Ensure `func` can actually process provided arguments. If `n_inputs` was
        not set, attempt to infer the number of required concurrent function
        calls from `args` and `kwargs`. In addition, ensure the size of each
        argument is "reasonable" for propagation across multiple workers.
        """

        # Ensure `func` really is a function and `n_inputs` makes sense
        if not callable(func):
            msg = "{} first input has to be a callable function, not {}"
            raise TypeError(msg.format(self.msgName, str(type(func))))
        msg = "{} `n_inputs` has to be 'auto' or an integer >= 2, not {}"
        if isinstance(n_inputs, str):
            if n_inputs != "auto":
                raise ValueError(msg.format(self.msgName, n_inputs))
            guessInputs = True
        else:
            try:
                acs._scalar_parser(n_inputs, varname="n_inputs", ntype="int_like", lims=[1, np.inf])
            except Exception as exc:
                raise exc
            guessInputs = False

        # Get `func`'s signature to extract its positional/keyword arguments
        funcSignature = inspect.signature(func)
        funcPosArgs = [name for name, value in funcSignature.parameters.items()\
            if value.default is value.empty and value.name != "kwargs"]
        funcKwargs = [name for name, value in funcSignature.parameters.items()\
            if value.default is not value.empty]

        # Account for positional args that were specified by name (keyword-like)
        args = list(args)
        posArgNames = []
        for name, value in kwargs.items():
            if name in funcPosArgs:
                args.insert(funcPosArgs.index(name), value)
                posArgNames.append(name)
        for name in posArgNames:
            kwargs.pop(name)

        # If "taskID" is a keyword arg, include/overwrite it in `kwargs` - the rest
        # is done by `ACMEdaemon`
        if "taskID" in funcKwargs:
            kwargs["taskID"] = None

        # Compare provided `args`/`kwargs` to actually defined quantities in `func`
        if len(args) != len(funcPosArgs):
            msg = "{} {} expects {} positional arguments ({}), found {}"
            validArgs = "'" + "'".join(arg + "', " for arg in funcPosArgs)[:-2]
            raise ValueError(msg.format(self.msgName,
                                        func.__name__,
                                        len(funcPosArgs),
                                        validArgs,
                                        len(args)))
        if len(kwargs) > len(funcKwargs):
            msg = "{} {} accepts at maximum {} keyword arguments ({}), found {}"
            validArgs = "'" + "'".join(arg + "', " for arg in funcKwargs)[:-2]
            raise ValueError(msg.format(self.msgName,
                                        func.__name__,
                                        len(funcKwargs),
                                        validArgs,
                                        len(kwargs)))

        # Prepare argument parsing: collect the the length of anything 1D-array-like
        # in `argLens` and check the size of all provided positional and keyword args
        argLens = []
        wrnMsg = "argument size {0:4.2f} MB exceeds recommended limit of {1} MB. " +\
            "Distributing large variables across workers may result in poor performance. "

        # Cycle through positional args
        for k, arg in enumerate(args):
            if isinstance(arg, range):
                arg = list(arg)
                args[k] = arg
            acs.callCount = 0
            argsize = acs.sizeOf(arg, "positional arguments")
            if argsize > self._maxArgSize:
                self.log.warning(wrnMsg.format(argsize, self._maxArgSize))
            if isinstance(arg, (list, tuple)):
                argLens.append(len(arg))
            elif isinstance(arg, np.ndarray):
                if len(arg.squeeze().shape) == 1:
                    argLens.append(arg.squeeze().size)

        # More complicated for keyword args: if provided value is 1D-array-like and
        # default value is not, interpret is as worker-arg list, and track its length
        for name, value in kwargs.items():
            defaultValue = funcSignature.parameters[name].default
            if isinstance(value, range):
                value = list(value)
                kwargs[name] = value
            acs.callCount = 0
            valsize = acs.sizeOf(value, "keyword arguments")
            if valsize > self._maxArgSize:
                self.log.warning(wrnMsg.format(valsize, self._maxArgSize))
            if isinstance(value, (list, tuple)):
                if isinstance(defaultValue, (list, tuple)):
                    if len(defaultValue) != len(value):
                        argLens.append(len(value))
                else:
                    argLens.append(len(value))
            elif isinstance(value, np.ndarray) and not isinstance(defaultValue, np.ndarray):
                if len(value.squeeze().shape) == 1:
                    argLens.append(value.squeeze().size)

        # If `n_input` is `"auto"`, make an educated guess as to how many parallel
        # executions of `func` are intended; if input args contained multiple
        # 1D-array-likes, pump the brakes. If `n_input` was explicitly provided,
        # ensure at least one input argument actually contains `n_input` elements
        # for distribution across parallel workers
        if guessInputs:
            if len(set(argLens)) > 1:
                msg = "{} automatic input distribution failed: found {} objects " +\
                    "containing {} to {} elements. Please specify `n_inputs` manually. "
                raise ValueError(msg.format(self.msgName, len(argLens), min(argLens), max(argLens)))
            n_inputs = argLens[0]
        else:
            if n_inputs not in set(argLens):
                msg = "{} No object has required length of {} matching `n_inputs`. "
                raise ValueError(msg.format(self.msgName, n_inputs))
        self.n_inputs = int(n_inputs)

        # Anything that does not contain `n_input` elements is converted to a one-element list
        wrnMsg = "Found a single callable object in positional arguments. " +\
            "It will be executed just once and shared by all workers"
        self.argv = list(args)
        for ak, arg in enumerate(args):
            if isinstance(arg, (list, tuple)):
                if len(arg) == self.n_inputs:
                    continue
            elif isinstance(arg, np.ndarray):
                if len(arg.squeeze().shape) == 1 and arg.squeeze().size == self.n_inputs:
                    continue
            elif callable(arg):
                self.log.warning(wrnMsg)
            self.argv[ak] = [arg]

        # Same for keyword arguments with the caveat that default values have to
        # be taken into account (cf. above)
        wrnMsg = "Found a single callable object in keyword arguments: {}. " +\
            "It will be executed just once and shared by all workers"
        self.kwargv = dict(kwargs)
        for name, value in kwargs.items():
            if isinstance(value, (list, tuple)):
                if len(value) == self.n_inputs:
                    continue
            elif isinstance(value, np.ndarray) and \
                not isinstance(funcSignature.parameters[name].default, np.ndarray):
                if len(value.squeeze().shape) == 1 and value.squeeze().size == self.n_inputs:
                    continue
            elif callable(value):
                self.log.warning(wrnMsg.format(name))
            self.kwargv[name] = [value]

        # Finally, attach user-provided function to class instance
        self.func = func

    def compute(self):
        """
        Shortcut to launch parallel computation via `ACMEdaemon`
        """
        if hasattr(self, "daemon"):
            self.daemon.compute()

    def cleanup(self):
        """
        Shortcut to corresponding cleanup-routine provided by `ACMEdaemon`
        """
        if hasattr(self, "daemon"):
            self.daemon.cleanup

    def __enter__(self):
        """
        If `ParallelMap` is used as context manager, launch `ACMEdaemon`
        """
        return self.daemon

    def __exit__(self, exception_type, exception_value, exception_traceback):
        """
        If `ParallelMap` is used as context manager, close any ad-hoc computing
        clients created by `ACMEdaemon`
        """
        self.daemon.cleanup()



# -*- coding: utf-8 -*-
#
# Simple script for testing acme w/o pip-installing it
#

# Builtin/3rd party package imports
import numpy as np

# Add acme to Python search path
import os
import sys
acme_path = os.path.abspath(".." + os.sep + "..")
if acme_path not in sys.path:
    sys.path.insert(0, acme_path)

# Import package
from acme import ParallelMap

def f(x, y, z=3, w=np.zeros((3, 1)), **kwargs):
    return (sum(x) + y) * z * w.max()


# Prepare code to be executed using, e.g., iPython's `%run` magic command
if __name__ == "__main__":

    # Test stuff within here...
    # pass

    pmap = ParallelMap(f, [2, 4, 6, 8], [2, 2], z=np.array([1, 2]), w=np.ones((8, 1)), n_inputs=2, dryrun=True)
    with pmap as p:
        p.compute()


